/* yamm -- yet another movemail, moves file locking it the way you
   want. Useful on systems where two locking techniques are available:
   flock or fcntl (namely Linux).
   Copyright (C) 1999 Alexander Vorobiev 
   $Id: yamm.c,v 1.1.1.1 1999/01/28 15:36:13 sparrow Exp $
*/

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this software; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
   Boston, MA 02111-1307 USA 
*/

/* Given no arguments yamm moves user's mailbox (specified by MAIL 
   environment variable) to standard output locking source file POSIX
   way i.e. using fcntl if availabe. 

   Command line arguments:

   Most general form is 
   yamm [<options>] [[<infilename>] <outfilename>]

   If <infilename> is missing, value specified by MAIL environment
   variable is used.
   If <outfilename> is missing, standard output is used.

   Options:

   --posix-locks or -p  use POSIX (fcntl) style of locking if available
   --bsd-locks   or -b  use BSD (flock) style of locking if available
   --help        or -h  display help 
   --version     or -v  output version information
*/


#include "yamm.h"

enum locktype_t {BSD_LOCK, POSIX_LOCK, lock_type_count};

enum locktype_t lock_type = POSIX_LOCK;  /* Use POSIX locks by default */

#define MAILBOX "MAIL"                 /* Name of environment variable */
#define BUFFSIZE 8192            /* Size of the buffer used for copying*/

#define ERR_MAILBOX_ENV "No MAIL environment variable found"
#define ERR_FLOCK       "- flock error"
#define ERR_FCNTL       "- fcntl error"
#define ERR_FTRUNCATE   "- ftruncate error"
#define ERR_WRITE       "- write error"
#define ERR_READ        "- read error"
#define ERR_NO_FCNTL    "This system does not support POSIX-style locks"
#define ERR_NO_FLOCK    "This system does not support BSD-style locks"
					  
/* Print help message to standard error and exit */

void exit_with_help()
{
  puts("\nyamm [<options>] [[<infilename>] <outfilename>]\n\n"
       "If <infilename> is missing, value specified by MAIL environment\n"
       "variable is used.\n"
       "If <outfilename> is missing, standard output is used.\n\n"
       "Options:\n\n"
       "-p, --posix-locks  use POSIX (fcntl) style of locking if available\n"
       "-b, --bsd-locks    use BSD (flock) style of locking if available\n"
       "-h, --help         display this help text and exit\n"
       "-v, --version      output version information and exit\n\n"
       "Given no arguments yamm moves user's mailbox (specified by MAIL\n"
       "environment variable) to standard output locking source file POSIX\n"
       "way i.e. using fcntl if availabe.\n");
  exit(0);
}

/* Prints program name and msg on standard error and optionally system 
 error message in between if sys_err set to nonzero value which is
 supposed to contain appropriate errno */

void exit_with_error(int sys_err, const char *msg)
{
  fprintf(stderr, "yamm: %s %s\n", 
	  sys_err ? (char *)strerror(sys_err) : "", msg);
  exit(1);
}

void display_version()
{
  printf("yamm %s\n", VERSION); /* VERSION is defined in config.h */
  exit(0);
}

/* Process options supplied. Returns 1 if locking style was set, 0
   otherwise */

int process_options(int argc, char **argv)
{
  int getopt_result;

  while(1) {
    static struct option long_options[] = {
      {"posix-locks", no_argument, NULL, 'p'},
      {"bsd-locks",   no_argument, NULL, 'b'},
      {"help",        no_argument, NULL, 'h'},
      {"version",     no_argument, NULL, 'v'},
      {0, 0, 0, 0}
    };

    getopt_result = getopt_long (argc, argv, "+pbhv", long_options, NULL);

    if (getopt_result == EOF)   /* All options have been processed */
      break;

    switch (getopt_result) {
    case 0:
      /* Nothing to do - lock_type is already set */
      break;
    case 'p':
      lock_type = POSIX_LOCK;
      break;
    case 'b':
      lock_type = BSD_LOCK;
      break;
    case 'v':
      display_version();
      break;
    default:
      return 0;
    };
  };
  return 1;
};

int open_file(char *filename)
{
  int fd;

  umask(0);
  if ((fd = open(filename, O_RDWR | O_CREAT, 
		 S_IRUSR | S_IWUSR | S_IRGRP | 
		 S_IWGRP | S_IROTH | S_IWOTH)) == -1)
    exit_with_error(errno, filename);
  else 
    return fd;
}

int open_mailbox(void)
{
  char *mailbox_file = getenv(MAILBOX);
  if (!mailbox_file)
    exit_with_error(0, ERR_MAILBOX_ENV);
  else 
    return open_file(mailbox_file);
}


/* (Un)locking functions */

void POSIX_lock(int fd)
{
#ifdef HAVE_FCNTL

  struct flock lock;
  
  lock.l_type   = F_WRLCK;
  lock.l_start  = 0;
  lock.l_whence = SEEK_SET;
  lock.l_len    = 0;

  if (fcntl(fd, F_SETLKW, &lock) == -1)
    exit_with_error(errno, ERR_FCNTL);

#else  /* not HAVE_FCNTL */

  exit_with_error(0, ERR_NO_FCNTL);

#endif /* not HAVE_FCNTL */
}

void POSIX_unlock(int fd)
{
#ifdef HAVE_FCNTL

  struct flock lock;
  
  lock.l_type   = F_UNLCK;
  lock.l_start  = 0;
  lock.l_whence = SEEK_SET;
  lock.l_len    = 0;

  if (fcntl(fd, F_SETLK, &lock) == -1)
    exit_with_error(errno, ERR_FCNTL);

#endif /* HAVE_FCNTL */
}

void BSD_lock(int fd)
{
#ifdef HAVE_FLOCK

  if (flock(fd, LOCK_EX))
    exit_with_error(errno, ERR_FLOCK);

#else  /* not HAVE_FLOCK */

  exit_with_error(0, ERR_NO_FLOCK);

#endif /* not HAVE_FLOCK */
}

void BSD_unlock(int fd)
{
#ifdef HAVE_FLOCK

  if (flock(fd, LOCK_UN) == -1)
    exit_with_error(errno, ERR_FLOCK);

#endif /* HAVE_FLOCK */
}

void copy_file(int infd, int outfd)
{
  char buf[BUFFSIZE];
  int  n;

  while ((n = read(infd, buf, BUFFSIZE)) > 0)
    if (write(outfd, buf, n) != n)
      exit_with_error(errno, ERR_WRITE);

  if (n < 0)
    exit_with_error(errno, ERR_READ);
}

int main(int argc, char **argv)
{
  int infd;  
  int outfd;

  void (*(lock_fun[lock_type_count]))(int);
  void (*(unlock_fun[lock_type_count]))(int);

  /* Process options */
  if (!process_options(argc, argv))
      exit_with_help();


  /* Get filenames if any and open the files */

  switch (argc - optind) {
  case 0:   /* No filenames - use default values */
    infd = open_mailbox();
    outfd = STDOUT_FILENO;
    break;
  case 1:   /* Output file name */
    infd = open_mailbox();
    outfd = open_file(argv[optind]);
    break;
  case 2:   /* Both input and output file names */
    infd = open_file(argv[optind]);
    outfd = open_file(argv[optind + 1]);
    break;
  default:
    exit_with_help();
  }

  /* Arrange (un)locking functions into arrays */

  lock_fun[POSIX_LOCK]   = &POSIX_lock;
  unlock_fun[POSIX_LOCK] = &POSIX_unlock;
  lock_fun[BSD_LOCK]     = &BSD_lock;
  unlock_fun[BSD_LOCK]   = &BSD_unlock;

  /* Lock input file the way chosen */

  (*(lock_fun[lock_type]))(infd);

  /* Move the file */

  copy_file(infd, outfd);

  if (ftruncate(infd, 0) == -1)
    exit_with_error(errno, ERR_FTRUNCATE);
  
  /* Unlock input file */

  (*(unlock_fun[lock_type]))(infd);

  return 0;
};
